#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>
#include <sys/mount.h>

#define CAPTURE_STDOUT "/tmp/stdout"
#define CAPTURE_STDERR "/tmp/stderr"

#define MAX_MODULENAME_LENGTH 15
#define MAX_CMDLINE_LENGTH 256
#define KERNEL_REAL_ROOT_DEV_FILE "/proc/sys/kernel/real-root-dev"
#define KERNEL_CMDLINE "/proc/cmdline"

int exec_insmod(char* modulestr);
int sys(const char *command, const char* parameter,
					const char *stdout_target, const char *stdout_mode,
					const char *stderr_target, const char *stderr_mode);

int main(int argc, char **argv, char **env)
{
  char *boot_modules, *real_root_dev, *root_dev_start, root_dev[15],
       cmdline[MAX_CMDLINE_LENGTH], module[80];
  int i=0, len, j;
  FILE *kernel_root_dev, *kernel_cmdline;
   
  boot_modules = getenv("boot_modules");
  real_root_dev = getenv("real_root_dev"); 
  if (boot_modules) {
    /* set the path where insmod searches for module files */
    setenv("MODPATH", "/lib", 1);

    /* set output streams to be unbuffered */
    setvbuf(stdout, (char*) NULL, _IONBF, 0);
    setvbuf(stderr, (char*) NULL, _IONBF, 0);
    fflush(stdout);
    fflush(stderr);

    printf("\nLinuxrc boot modules loader starting\n");
    printf("Rene Mayrhofer <rmayr@vianova.at>, Sep. 1999\n\n");
    
    /* parse the boot_modules string */
    len=strlen(boot_modules);
    while(i < len) {
      /* get the modules to load (separated by commas) */
      j=0;
      while(i < len && boot_modules[i] != ',') {
        module[j] = boot_modules[i];
				i++; j++;
      }
      module[j] = 0;
      if(j > 0) {
				exec_insmod(module);
      }
      i++;
    }
  } else {
    printf("Warning: boot_modules variable not found in kernel parameters.\n");
  }
  fputs("\n", stdout);
   
  /* set the real root device (needed if the root dev was not detected 
     by the kernel during initial bootup) */
  /* first mount the proc dev to be able to manipulate the kernel */
  if(mount("proc", "/proc", "proc", 0, 0) != 0) {
    printf("Unable to mount the proc filesystem !\n");
    exit(2);
  }
  /* parse the kernel commandline for an occurance of a root statement */
  if( (kernel_cmdline = fopen(KERNEL_CMDLINE, "r")) == NULL) {
    printf("Unable to open %s !\n", KERNEL_CMDLINE);
    exit(3);
  }
  if(fgets(cmdline, MAX_CMDLINE_LENGTH, kernel_cmdline) == NULL) {
    printf("Unable to read from %s !\n", KERNEL_CMDLINE);
    exit(4);
  }
  fclose(kernel_cmdline);
  if( (root_dev_start = strstr(cmdline, "root=")) != NULL) {
    len = strlen(root_dev_start);
    i = 5;
    while(i < len && root_dev_start[i] != ' ') {
      root_dev[i-5] = root_dev_start[i];
      i++;
    }
    root_dev[i-5] = 0;
    printf("Root device as set on the kernel command line: %s\n", root_dev);
  } else {
    printf("Warning: Could not find a root definition on kernel command line.\n"); 
  }
  
  /* if the real root device was given on the kernel command line, use it */
  if(real_root_dev) {
    printf("Setting kernel root device to %s.... ", real_root_dev);
    if( (kernel_root_dev = fopen(KERNEL_REAL_ROOT_DEV_FILE, "w")) == NULL) {    
      printf("Unable to open %s !\n", KERNEL_REAL_ROOT_DEV_FILE);
      exit(5);
    }
    if(fputs(real_root_dev, kernel_root_dev) == EOF) {
      printf("Unable to set real root device !\n");
      exit(6);
    }
    fclose(kernel_root_dev); 
    printf("successful.\n");
  }
   
  if(umount("/proc") != 0) {
    printf("Unable to unmount the proc filesystem !\n");
  }
  
  fputs("\n",stdout);
  return 0;
}

int exec_insmod(char *modulestr) {
  FILE *read_stderr;
  int i=0, len, ret, ch, name_ended=0, name_len=0;
  char modulename[MAX_MODULENAME_LENGTH];
   
  len=strlen(modulestr);
  /* get the module parameters (separated by ':' */
  while(i < len) {
    if(modulestr[i] == ':') {
      modulestr[i] = ' ';
      if(! name_ended)
        name_ended = 1;
    }
    else {
      if(! name_ended) {
        modulename[i] = modulestr[i];
        name_len++;
      }
    }
    i++;
  }
  modulename[name_len] = 0;
  printf("loading module: %-*s.... ", MAX_MODULENAME_LENGTH, modulename);
  
  /* try to execute insmod */
	ret = sys("/sbin/insmod", modulestr, CAPTURE_STDOUT, "a", CAPTURE_STDERR, "w");
   
  if(ret == 0) {
    printf("successfully loaded.\n");
    return 0;
  } else {
    printf("unsuccessful ! The error was:\n");
    read_stderr = fopen(CAPTURE_STDERR, "r");
    while( (ch = fgetc(read_stderr)) != EOF)
      putc(ch, stdout);
    fclose(read_stderr);
    printf("\n");
    return 2;
  }
  
  return 3;
}


/* mainly taken from "man 3 system" */
int sys(const char *command, const char *parameter,
					const char *stdout_target, const char *stdout_mode,
					const char *stderr_target, const char *stderr_mode) {
	int pid, status;
	char *argv[4];

	if(command == 0)
		return 1;
	/* start a new process */
	pid = fork();
	if(pid == -1)
		return -1;
	if(pid == 0) {
		argv[0] = command;
		argv[1] = parameter;
		argv[2] = 0;
		
		/* set the output places */
	  /* (set stdout and stderr to temporary files to capture the output) */
	  if( (stdout = freopen(stdout_target, stdout_mode, stdout)) == NULL) {
  	  return -2;
	  }
	  if( (stderr = freopen(stderr_target, stderr_mode, stderr)) == NULL) {
	    return -2;
	  }
		
		/* execute */
		execv(command, argv);
		exit(127);
	}
	
	/* wait for completion */
	do {
		if(waitpid(pid, &status, 0) == -1) {
			if(errno != EINTR) {
				return -1;
			}
		} else
			return status;
	} while(1);
}
